@@ -6,11 +6,11 @@ use std::{
     borrow::Cow,
     convert::TryInto,
     fmt::{self, Debug, Formatter},
-    io::Write,
+    io::{ErrorKind, Write},
     str::FromStr,
 };
 use trillium_http::{
-    Body, Error, HeaderName, HeaderValue, Headers,
+    Body, Error, HeaderName, HeaderValue, HeaderValues, Headers,
     KnownHeaderName::{
         Connection, ContentLength, Expect, Host, ProxyConnection, TransferEncoding, UserAgent,
     },
@@ -21,6 +21,21 @@ use url::Url;
 const MAX_HEADERS: usize = 128;
 const MAX_HEAD_LENGTH: usize = 2 * 1024;
 
+/**
+A wrapper error for [`trillium_http::Error`] or
+[`serde_json::Error`]. Only available when the `json` crate feature is
+enabled.
+*/
+#[cfg(feature = "json")]
+#[derive(thiserror::Error, Debug)]
+pub enum ClientSerdeError {
+    #[error(transparent)]
+    HttpError(#[from] Error),
+
+    #[error(transparent)]
+    JsonError(#[from] serde_json::Error),
+}
+
 /**
 a client connection, representing both an outbound http request and a
 http response
@@ -288,6 +303,42 @@ impl<C: Connector> Conn<'_, C> {
         &mut self.request_headers
     }
 
+    /**
+    chainable setter for [`inserting`](Headers::insert) a request header
+
+    ```
+    use trillium_smol::TcpConnector;
+    type Conn = trillium_client::Conn<'static, TcpConnector>;
+
+    let handler = |conn: trillium::Conn| async move {
+        let header = conn.headers().get_str("some-request-header").unwrap_or_default();
+        let response = format!("some-request-header was {}", header);
+        conn.ok(response)
+    };
+
+    trillium_testing::with_server(handler, |url| async move {
+        let mut conn = Conn::get(url)
+            .with_header("some-request-header", "header-value") // <--
+            .execute()
+            .await?;
+        assert_eq!(
+            conn.response_body().read_string().await?,
+            "some-request-header was header-value"
+        );
+        Ok(())
+    })
+    ```
+    */
+
+    pub fn with_header(
+        mut self,
+        name: impl Into<HeaderName<'static>>,
+        value: impl Into<HeaderValues>,
+    ) -> Self {
+        self.request_headers.insert(name, value);
+        self
+    }
+
     /**
     ```
     use trillium_smol::TcpConnector;
@@ -345,6 +396,11 @@ impl<C: Connector> Conn<'_, C> {
     }
 
     /**
+    chainable setter to assign the request body. deprecated in
+    preference to the renamed [`Conn::with_body`] since it wouldn't
+    make sense to set a response body on a client conn.
+
+
     ```
     env_logger::init();
     use trillium_smol::TcpConnector;
@@ -369,14 +425,54 @@ impl<C: Connector> Conn<'_, C> {
     });
     ```
     */
+    #[deprecated = "renamed as with_body. will be removed at the next minor release"]
     pub fn with_request_body(mut self, body: impl Into<Body>) -> Self {
         self.set_request_body(body);
         self
     }
 
-    // pub(crate) fn request_encoding(&self) -> &'static Encoding {
-    //     encoding(&self.request_headers)
-    // }
+    /**
+    chainable setter for the request body
+
+    ```
+    env_logger::init();
+    use trillium_smol::TcpConnector;
+    type Conn = trillium_client::Conn<'static, TcpConnector>;
+
+    let handler = |mut conn: trillium::Conn| async move {
+        let body = conn.request_body_string().await.unwrap();
+        conn.ok(format!("request body was: {}", body))
+    };
+
+    trillium_testing::with_server(handler, |url| async move {
+        let mut conn = Conn::post(url)
+            .with_body("body") //<-
+            .execute()
+            .await?;
+
+        assert_eq!(
+            conn.response_body().read_string().await?,
+            "request body was: body"
+        );
+        Ok(())
+    });
+    ```
+     */
+    pub fn with_body(mut self, body: impl Into<Body>) -> Self {
+        self.set_request_body(body);
+        self
+    }
+
+    /**
+    chainable setter for json body. this requires the `json` crate feature to be enabled.
+     */
+    #[cfg(feature = "json")]
+    pub fn with_json_body(self, body: &impl serde::Serialize) -> serde_json::Result<Self> {
+        Ok(self.with_body(serde_json::to_string(body)?).with_header(
+            trillium_http::KnownHeaderName::ContentType,
+            "application/json",
+        ))
+    }
 
     pub(crate) fn response_encoding(&self) -> &'static Encoding {
         encoding(&self.response_headers)
@@ -451,10 +547,23 @@ impl<C: Connector> Conn<'_, C> {
         )
     }
 
+    /**
+    Attempt to deserialize the response body. Note that this consumes the body content.
+     */
+    #[cfg(feature = "json")]
+    pub async fn response_json<T>(&mut self) -> std::result::Result<T, ClientSerdeError>
+    where
+        T: serde::de::DeserializeOwned,
+    {
+        let body = self.response_body().read_string().await?;
+        Ok(serde_json::from_str(&body)?)
+    }
+
     pub(crate) fn response_content_length(&self) -> Option<u64> {
-        if self.method == Method::Head {
-            Some(0)
-        } else if let Some(Status::NoContent | Status::NotModified) = self.status {
+        if self.status == Some(Status::NoContent)
+            || self.status == Some(Status::NotModified)
+            || self.method == Method::Head
+        {
             Some(0)
         } else {
             self.response_headers
@@ -561,7 +670,10 @@ impl<C: Connector> Conn<'_, C> {
 
     async fn connect_and_send_head(&mut self) -> Result<()> {
         if self.transport.is_some() {
-            panic!("cannot connect a second time");
+            return Err(Error::Io(std::io::Error::new(
+                ErrorKind::AlreadyExists,
+                "conn already connected",
+            )));
         }
 
         let head = self.build_head().await?;
